﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace PMDMikuDirect3d
{
    public partial class SimpleLiteD3d : IDisposable
    {
       public struct PMD_Header    //【PMDファイル ヘッダー情報】
       {  public string ID;        //【ID】3バイト固定長 "P","m","d"
          public float  vNum;      // バージョン番号
          public string ModelName; //【モデル名称】20バイト固定長
          public string CopyRight; //【著作権情報】256バイト固定長
       }

       public struct PMD_vertex     //【PMD_頂点データ】　4*3+4*3+4*2+2*2+2 = 12+12+8+4+2 = 38bytes/頂点
       {  public int nPoints;       //【頂点個数】 
          public CustomVertex.PositionNormalTextured[] UserVertex; //【頂点座標】【法線ﾍﾞｸﾄﾙ】【uv座標】
          public CustomVertex.PositionNormalTextured[] OriginalVertex;  //【頂点座標】【法線ﾍﾞｸﾄﾙ】【uv座標】SphereMap対応
          public short[] Bone1Num;  //【Bone番号】Bone番号1 モデル変形(頂点移動)時に影響
          public short[] Bone2Num;  //【Bone番号】Bone番号2      :
          public Byte[] BoneWeight; //【Bone重み係数】Bone1に与える影響度 0-100 Bone2への影響度は(100-bone_weight)で与えられる。
          public Byte[] EdgeFlag;   //【ｴｯｼﾞ･ﾌﾗｸﾞ】0:通常、1:エッジ無効　輪郭線が有効の場合
       }

       public struct PMD_index        //【PMD_インデックスデータ(面情報)】MikuMikuDanceのモデルではtriangle(三角形)しか使っていません。
       {  public int nIndex;          //【配列要素数】＝面数(三角形の個数)*3
          public int nPolygons;       //●三角形の個数
          public short[] IndexBuffer; //【頂点番号リスト】３個/面
       } //注）本質的にはGPU側の制約によるが、この配列のためにMMDでは最大65535頂点までのモデルしか取り扱えない。

       public struct PMD_material          //【PMD_材質データ】
       {  public int nMaterial;            //【材質個数】 
          public Material[] Material;      //【材質データ】DirectXの材質ﾃﾞｰﾀ構造体と同じ
          public byte[] ToonIndex;         //【ﾄｩｰﾝ･ｲﾝﾃﾞｯｸｽ番号】toon??.bmp 0.bmp:0xFF, 1.bmp:0x00, ･･･10.bmp:0x09
          public byte[] EdgeFlag;          //【ｴｯｼﾞ･ﾌﾗｸﾞ】輪郭､影
          public int[] nPoints;            //【面構成要素数】面数*3 = 面頂点数 実際のindexに直すには材質0から累積加算する。
          public string[] TextureFileName; //【ﾃｸｽﾁｬ･ﾌｧｲﾙ名】20bytesギリギリまで使える 20bytes時は0x00がなくても可。'
          public Texture[] Texture;        //●【ﾃｸｽﾁｬ】
          public bool[] SphereMap;         //●【SphereMapフラグ】
          public bool SphereFlag;          //●【描画時フラグ】Sphere無しなら高速に描画
       }

       public struct PMD_bone //【PMD_Boneデータ】
       {
           public short nBone;             //【Bone数】
           public string[] BoneName;       //【Bone名称】PMDファイルでは固定長(20bytes)
           public short[] ParentBoneIndex; //【親Bone頂点番号】(無い場合は0xFFFF)
           public short[] ChildBoneIndex;  //【子Bone頂点番号】(chain末端は0xFFFF)親:子は1:多なので位置決め用
           public byte[] BoneType;         //【Boneの種類】0:回転のみ 1:回転と移動 2:IK 4:影響下 5:回転影響下 6:IK接続先 8:捻り 9:回転運動
           public short[] Dummy;           //【不明】
           public Vector3[] BoneHead;      //【Boneのヘッド座標】(x,y,z)位置ベクトル
       }

       public struct PMD_IKChain
       {
           public short[] Index; //【IK影響下のBone番号】
       }

       public struct PMD_IKbone //【PMD_IKBoneデータ】
       {
           public short nIKBone;             //【IKBone数】
           public short[] IKBoneIndex;       //【IK Bone番号】
           public short[] IKTargetBoneIndex; //【IK接続先Bone番号】
           public byte[] IKChainLength;      //【IK鎖の長さ】子の数
           public short[] Iterations;        //【再帰演算回数】
           public float[] ControlWeight;     //【重み係数】IKの影響度
           public PMD_IKChain[] IKChildBoneIndex;  //【IK影響下のBone番号】
       }

       public struct PMD_Vectors //【PMD_表情頂点データ】
       {
           public int[] Index;        //【頂点番号】
           public Vector3[] Position; //【頂点位置ベクトルまたは表情ベクトル】
       }

       public struct PMD_Skin //【PMD_Skinデータ】
       {
           public short nSkin;                   //【Skin数】
           public string[] SkinName;             //【表情名】
           public int[] nSkinPoints;             //【表情用頂点数】
           public byte[] SkinType;               //【表情の種類】 0:base 1:眉 2:リップ 4:その他
           public PMD_Vectors[] SkinPointVector; //【表情用頂点座標】位置ベクトル
           public PMD_Vectors SkinSetData;       //●最終的にUserVertexBufferにセットする頂点座標
           public float[] SkinEffect;            //●各表情ベクトルの係数値(nSkin個)
       }

       public PMD_Header   PMDh; //【PMDヘッダー情報】
       public PMD_vertex   PMDv; //【PMD頂点バッファ】
       public PMD_index    PMDi; //【PMDインデックス】
       public PMD_material PMDm; //【PMD材質情報】
       public PMD_bone     PMDb; //【PMDボーン】
       public PMD_IKbone PMDIKb; //【PMDIKボーン】
       public PMD_Skin     PMDs; //【PMD表情】
       public string TextBuffer; //【テキストの読み込みに使うバッファ】
       public const float Zoom = 20.0f; //【変換倍率】座標系をMMD⇒ARに変換する際、1ミクセルを20倍にして表示。
       public const string PMDPath = "../../PMD/"; //【PMDファイルの相対パス】
//       public const string PMDPath = "c:/My3DCG/PMDデータ/LatMikuVer2.3/"; //【PMDファイルの相対パス】

       //***********************************************************************************************
       //*** FileStreamとBinaryReaderを使ってPMDファイルからMMD構造体に情報を読み込む
       //***********************************************************************************************
       public void ReadPMDFile(string PMDFilePath) //【MikuMikuDanceのPMDファイルからHeader, Vertex, Index, Meterialを読み込む】
       {  
          FileStream PMDFile = new FileStream(PMDFilePath, FileMode.Open); //【PMDファイル読み込みの為のFileStream】    
          BinaryReader bRD = new BinaryReader(PMDFile);                    //【PMDファイル読み込みの為のBinaryReader】     

          //【PMDヘッダー情報を読み込む】
          //●文字列は改行や空白等の不必要な領域も含めて読み込んでいる。
          PMDh.ID = Encoding.GetEncoding(932).GetString(bRD.ReadBytes(3));          //【IDを読み込む】固定長３バイト "P","m","d"
          PMDh.vNum = bRD.ReadSingle();                                             //【バージョン番号を読み込む】
          PMDh.ModelName = Encoding.GetEncoding(932).GetString(bRD.ReadBytes(20));  //【モデル名称を読み込む】固定長20バイト
          PMDh.CopyRight = Encoding.GetEncoding(932).GetString(bRD.ReadBytes(256)); //【著作権情報を読み込む】固定長256バイト

           //【PMD頂点情報を読み込む】
          PMDv.nPoints = bRD.ReadInt32(); //
          PMDv.UserVertex = new CustomVertex.PositionNormalTextured[PMDv.nPoints]; // Tu,TvはSphereMapで変更されることがある。
          PMDv.OriginalVertex = new CustomVertex.PositionNormalTextured[PMDv.nPoints];  //【Original】
          PMDv.Bone1Num = new short[PMDv.nPoints];  //【Bone番号】Bone番号1 モデル変形(頂点移動)時に影響
          PMDv.Bone2Num = new short[PMDv.nPoints];  //【Bone番号】Bone番号2      :
          PMDv.BoneWeight = new Byte[PMDv.nPoints]; //【Bone重み係数】Bone1に与える影響度 0-100 Bone2への影響度は(100-bone_weight)で与えられる。
          PMDv.EdgeFlag = new Byte[PMDv.nPoints];   //【ｴｯｼﾞ･ﾌﾗｸﾞ】0:通常、1:エッジ無効　輪郭線が有効の場合
          for (var i = 0; i < PMDv.nPoints; i++)    //【右手系から左手系に】Z⇒-Z, Nz⇒-Nz
          {  PMDv.UserVertex[i].Position = new Vector3(bRD.ReadSingle()*Zoom, bRD.ReadSingle()*Zoom, -bRD.ReadSingle()*Zoom); //【位置ベクトル】X,Y,Z
             PMDv.UserVertex[i].Normal   = new Vector3(bRD.ReadSingle(), bRD.ReadSingle(), -bRD.ReadSingle()); //【法線ベクトル】Nx, Ny, Nz
             PMDv.UserVertex[i].Tu       = bRD.ReadSingle(); //【テクスチャーUV空間】u
             PMDv.UserVertex[i].Tv       = bRD.ReadSingle(); //【テクスチャーUV空間】v
             PMDv.Bone1Num[i]   = bRD.ReadInt16();
             PMDv.Bone2Num[i]   = bRD.ReadInt16();
             PMDv.BoneWeight[i] = bRD.ReadByte();
             PMDv.EdgeFlag[i]   = bRD.ReadByte();
             PMDv.OriginalVertex[i] = PMDv.UserVertex[i]; //【ShpereMapでTu, Tvを壊すのでオリジナルを保存しておく】
          }

           //【PMD面(インデックス)情報を読み込む】
          PMDi.nIndex = bRD.ReadInt32();
          PMDi.nPolygons = PMDi.nIndex / 3;
          PMDi.IndexBuffer = new short[PMDi.nIndex];
          for (var i = 0; i < PMDi.nIndex; i++) { PMDi.IndexBuffer[i] = bRD.ReadInt16(); }

           //【PMD材質情報を読み込む】 
          PMDm.nMaterial = bRD.ReadInt32();
          PMDm.Material = new Material[PMDm.nMaterial];
          PMDm.ToonIndex = new byte[PMDm.nMaterial];
          PMDm.EdgeFlag = new byte[PMDm.nMaterial];
          PMDm.nPoints = new int[PMDm.nMaterial];
          PMDm.TextureFileName = new string[PMDm.nMaterial];
          PMDm.SphereMap = new bool[PMDm.nMaterial];
          PMDm.SphereFlag = false;
          for (var i = 0; i < PMDm.nMaterial; i++)
          {  PMDm.Material[i].AmbientColor = new ColorValue(bRD.ReadSingle(), bRD.ReadSingle(), bRD.ReadSingle(), bRD.ReadSingle()); //【環境色】Red,Green,Blue,Alpha
             PMDm.Material[i].SpecularSharpness = bRD.ReadSingle();                                                 //【鏡面反射強度】
             PMDm.Material[i].SpecularColor = new ColorValue(bRD.ReadSingle(), bRD.ReadSingle(), bRD.ReadSingle()); //【鏡面反射色】Red, Green, Blue 
             PMDm.Material[i].DiffuseColor  = new ColorValue(bRD.ReadSingle(), bRD.ReadSingle(), bRD.ReadSingle()); //【拡散光色】　Red, Green, Blue
             PMDm.Material[i].EmissiveColor = new ColorValue(0.0F, 0.0F, 0.0F); //【発光色はMikuMikuDanceでは未使用】
             PMDm.ToonIndex[i] = bRD.ReadByte();
             PMDm.EdgeFlag[i] = bRD.ReadByte();
             PMDm.nPoints[i] = bRD.ReadInt32();
             PMDm.TextureFileName[i] = ""; //【テクスチャーファイル名読み込み】固定長20バイト
             TextBuffer = Encoding.GetEncoding(932).GetString(bRD.ReadBytes(20));
             for (var k = 0; k < 20; k++)  //【nullコード直前まで読み込む】
             {  if (TextBuffer[k] == 0) { break; } else { PMDm.TextureFileName[i] += TextBuffer[k]; }
             }
             PMDm.SphereMap[i] = false;
             if (PMDm.TextureFileName[i] != "")
             {
                 string[] p = PMDm.TextureFileName[i].Split(new char[] { '.' });
                 if (p[1] == "sph") { PMDm.SphereMap[i] = true; PMDm.SphereFlag = true;  } //【ファイル名がsph形式の場合】
                 PMDm.TextureFileName[i] = PMDPath + PMDm.TextureFileName[i]; 
             }
          }

          //【PMDボーン情報を読み込む】
          PMDb.nBone = bRD.ReadInt16();
          PMDb.BoneName = new string[PMDb.nBone];       //【Bone名称】20bytes
          PMDb.ParentBoneIndex = new short[PMDb.nBone]; //【親Bone頂点番号】(無い場合は0xFFFF)
          PMDb.ChildBoneIndex = new short[PMDb.nBone];  //【子Bone頂点番号】(chain末端は0xFFFF)親:子は1:多なので位置決め用
          PMDb.BoneType = new byte[PMDb.nBone];         //【Boneの種類】0:回転のみ 1:回転と移動 2:IK 4:影響下 5:回転影響下 6:IK接続先 8:捻り 9:回転運動 
          PMDb.Dummy = new short[PMDb.nBone];           //【不明】IK接続先？
          PMDb.BoneHead = new Vector3[PMDb.nBone];      //【Boneのヘッド座標】(x,y,z)位置ベクトル
          for (var i = 0; i < PMDb.nBone; i++)
          {  PMDb.BoneName[i] = ""; //【ボーン名読み込み】固定長20バイト
             TextBuffer = Encoding.GetEncoding(932).GetString(bRD.ReadBytes(20));
             for (var k = 0; k < 20; k++)  //【nullコード直前まで読み込む】
             {  if (TextBuffer[k] == 0) { break; } else { PMDb.BoneName[i] += TextBuffer[k]; }
             }
             PMDb.ParentBoneIndex[i] = bRD.ReadInt16();
             PMDb.ChildBoneIndex[i] = bRD.ReadInt16();
             PMDb.BoneType[i] = bRD.ReadByte();
             PMDb.Dummy[i] = bRD.ReadInt16();
             PMDb.BoneHead[i] = new Vector3(bRD.ReadSingle(), bRD.ReadSingle(), bRD.ReadSingle());
          }

          //【PMD IKボーン情報を読み込む】
          PMDIKb.nIKBone = bRD.ReadInt16(); //【IKBone数】2bytes? 1bytesでは？
          PMDIKb.IKBoneIndex = new short[PMDIKb.nIKBone];       //【IK Bone番号】
          PMDIKb.IKTargetBoneIndex = new short[PMDIKb.nIKBone]; //【IK接続先Bone番号】
          PMDIKb.IKChainLength = new byte[PMDIKb.nIKBone];      //【IK鎖の長さ】子の数
          PMDIKb.Iterations = new short[PMDIKb.nIKBone];        //【再帰演算回数】
          PMDIKb.ControlWeight = new float[PMDIKb.nIKBone];     //【重み係数】IKの影響度
          PMDIKb.IKChildBoneIndex = new PMD_IKChain[PMDIKb.nIKBone];  //【IK影響下のBone番号】IKChainLength=0ならなし
          for (var i = 0; i < PMDIKb.nIKBone; i++)
          {   PMDIKb.IKBoneIndex[i] = bRD.ReadInt16();
              PMDIKb.IKTargetBoneIndex[i] = bRD.ReadInt16();
              PMDIKb.IKChainLength[i] = bRD.ReadByte();
              PMDIKb.Iterations[i] = bRD.ReadInt16();
              PMDIKb.ControlWeight[i] = bRD.ReadSingle();

              if (PMDIKb.IKChainLength[i] != 0)
              {  PMDIKb.IKChildBoneIndex[i].Index = new short[PMDIKb.IKChainLength[i]];
                 for (var j = 0; j < PMDIKb.IKChainLength[i]; j++)
                 {  PMDIKb.IKChildBoneIndex[i].Index[j] = bRD.ReadInt16();
                 }
              }
          }

          //【PMDスキン情報を読み込む】
          PMDs.nSkin = bRD.ReadInt16();                       //【Skin数】
          PMDs.SkinName = new string[PMDs.nSkin];             //【表情名】
          PMDs.nSkinPoints = new int[PMDs.nSkin];             //【表情用頂点数】
          PMDs.SkinType = new byte[PMDs.nSkin];               //【表情の種類】 0:base 1:眉 2:リップ 4:その他
          PMDs.SkinPointVector = new PMD_Vectors[PMDs.nSkin]; //【表情用頂点座標】位置ベクトル
          PMDs.SkinEffect = new float[PMDs.nSkin];            //【各表情ベクトルの係数】0.0～1.0
          for (var i = 0; i < PMDs.nSkin; i++)
          {  PMDs.SkinName[i] = "";                    //【表情名】固定長20バイト
             TextBuffer = Encoding.GetEncoding(932).GetString(bRD.ReadBytes(20));
             for (var k = 0; k < 20; k++)  //【nullコード直前まで読み込む】
             {  if (TextBuffer[k] == 0) { break; } else { PMDs.SkinName[i] += TextBuffer[k]; }
             }
             PMDs.nSkinPoints[i] = bRD.ReadInt32();    //【表情用頂点数】
             PMDs.SkinType[i] = bRD.ReadByte();        //【表情の種類】 0:base 1:眉 2:リップ 4:その他
             PMDs.SkinPointVector[i].Index = new int[PMDs.nSkinPoints[i]];
             PMDs.SkinPointVector[i].Position = new Vector3[PMDs.nSkinPoints[i]];
             for (var j = 0; j < PMDs.nSkinPoints[i]; j++)
             {  PMDs.SkinPointVector[i].Index[j] = bRD.ReadInt32(); //【表情用頂点インデックス】
                //【表情用頂点位置ベクトル/表情ベクトル】表示倍率Zoomをかけておく。Zは頂点と同様反転するので注意。 
                PMDs.SkinPointVector[i].Position[j] = new Vector3(Zoom * bRD.ReadSingle(), Zoom * bRD.ReadSingle(), -Zoom * bRD.ReadSingle());
             }
             PMDs.SkinEffect[i] = 0.0f; //【各表情ベクトル係数の初期化】
          }
          PMDs.SkinSetData.Index = new int[PMDs.nSkinPoints[0]];        //【表情設定用Indexデータ】表情ベクトルを設定するVertexのIndex値格納用
          PMDs.SkinSetData.Position = new Vector3[PMDs.nSkinPoints[0]]; //【表情設定用Vertexデータ】各表情ベクトルを加算したVertex値格納用
          for (var i = 0; i < PMDs.nSkinPoints[0]; i++) //【表情ベクトルの頂点番号リストの初期化】Indexは１回コピーしておけばよい
          {  PMDs.SkinSetData.Index[i] = PMDs.SkinPointVector[0].Index[i];  }    

          //【PMDファイルを閉じる】
          bRD.Close();
          PMDFile.Close();
       }
    
       //***********************************************************************************************
       //*** 表情ベクトルの初期化を行う
       //***********************************************************************************************
       public void InitializeFacialExpression()
       {  for (var i = 0; i < PMDs.nSkinPoints[0]; i++) //【表情ベクトル頂点座標初期化】
          { PMDs.SkinSetData.Position[i] = PMDs.SkinPointVector[0].Position[i]; }
       }

       //***********************************************************************************************
       //*** 表情名称で表情ベクトルを探し、発見できればbaseに加算する
       //***********************************************************************************************
       public void SetFacialExpression(string SkinName, float SkinEffect)
       {  int SkinNumber = 0;
          for (var i = 1; i < PMDs.nSkin; i++)  {  if (PMDs.SkinName[i] == SkinName) { SkinNumber = i; }  }
          if (SkinNumber != 0) //【同じ名称の表情が見つかった場合、表情ベクトルをbaseに加える】
          {  for (var i = 0; i < PMDs.nSkinPoints[SkinNumber]; i++) //
              { PMDs.SkinSetData.Position[PMDs.SkinPointVector[SkinNumber].Index[i]] += SkinEffect * PMDs.SkinPointVector[SkinNumber].Position[i]; }
          }
       }

       //***********************************************************************************************
       //*** 指定表情番号の表情ベクトルをbaseに加算する
       //***********************************************************************************************
       public void SetFacialExpression(int SkinNumber, float SkinEffect)
       {  if (SkinNumber > 0 && SkinNumber < PMDs.nSkin) //【０はbaseなので不可】
          {  for (var i = 0; i < PMDs.nSkinPoints[SkinNumber]; i++) //
             { PMDs.SkinSetData.Position[PMDs.SkinPointVector[SkinNumber].Index[i]] += SkinEffect * PMDs.SkinPointVector[SkinNumber].Position[i]; }
          }
       }

       //***********************************************************************************************
       //*** 表情ベクトルをUserVertexBufferに設定する
       //***********************************************************************************************
       public void TransferFacialExpression()
       {  for (var i=0; i < PMDs.nSkinPoints[0]; i++)
          {  PMDv.UserVertex[PMDs.SkinSetData.Index[i]].Position = PMDs.SkinSetData.Position[i]; }
       }
    }
}
